/*  ***** BEGIN LICENSE BLOCK *****
 *  Version: RPPL 1.0/GPL 2.0/LGPL 2.1/MPL 1.1/CDDL 1.0
 *  
 *  
 *  The contents of this file are subject to the Roland Pihlakas Public
 *  License Version 1.0 (the "License"); you may not use this file except
 *  in compliance with the License. You may obtain a copy of the License
 *  at http://roland.pri.ee/RPPL/
 *  
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *  
 *  The Original Code is: Modelling of Natural Thinking.
 *  
 *  The Initial Developer of the Original Code is: Roland Pihlakas.
 *  
 *  Portions created by: _____________________________ 
 *  are Copyright (C): _____________________________. 
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 *  
 *  Alternatively, the contents of this file may be used under the terms of 
 *  either the GNU General Public License Version 2 or later (the "GPL"), or 
 *  the GNU Lesser General Public License Version 2.1 or later (the "LGPL"), or 
 *  the Mozilla Public Licence Version 1.1 or later (the "MPL"), or the Common 
 *  Development and Distribution License Version 1.0 or later (the "CDDL"), in 
 *  which case the provisions of the MPL, the GPL, the LGPL or the CDDL are 
 *  applicable instead of those above. If you wish to allow use of your version 
 *  of this file only under the terms of the of either the MPL, the GPL, 
 *  the LGPL or the CDDL, and not to allow others to use your version of this 
 *  file under the RPPL, indicate your decision by deleting the provisions 
 *  above and replace them with the notice and other provisions required by 
 *  the MPL, the GPL, the LGPL or the CDDL. If you do not delete the provisions 
 *  above, a recipient may use your version of this file under the terms of any 
 *  one of the RPPL, the MPL, the GPL, the LGPL or the CDDL.
 *
 *  ***** END LICENSE BLOCK ***** */

//######################################################################################################
//
//######################################################################################################



///#######################################################################
///
///#######################################################################

template <int size, typename functor, typename blocktype = _ss_block, int index = 0>
struct _conv_mm_xs_worker
{
	static inline void apply(blocktype const (&a_in)[size], blocktype (&a_out)[size / 2])
	{
		_mm_xs_worker1<functor, blocktype>::apply(a_out[index / 2], a_in[index], a_in[index + 1]);	
				
		_conv_mm_xs_worker<size, functor, blocktype, index + 2>::apply(a_in, a_out);
	};
};

template <int size, typename functor, typename blocktype>
struct _conv_mm_xs_worker<size, functor, blocktype, size>					///end loop
{
	static inline void apply(blocktype const (&UNUSED(a_in))[size], blocktype (&UNUSED(a_out))[size / 2]) {};
};

///#######################################################################
///
///#######################################################################

template <bool nextlevel_has_remainder, bool even>
struct _get_conv_mm_xs_inner2_has_remainder2;


template <>
struct _get_conv_mm_xs_inner2_has_remainder2<false, false>
{
	static const bool rval = true;
};

template <>
struct _get_conv_mm_xs_inner2_has_remainder2<false, true>
{
	static const bool rval = false;
};

template <>
struct _get_conv_mm_xs_inner2_has_remainder2<true, false>
{
	static const bool rval = false;
};

template <>
struct _get_conv_mm_xs_inner2_has_remainder2<true, true>
{
	static const bool rval = true;
};

///#######################################################################

template <int size>
struct _get_conv_mm_xs_inner2_has_remainder
{
	static const bool nextlevel_has_remainder = _get_conv_mm_xs_inner2_has_remainder<size / 2>::rval;
	static const bool even = ((size & 0x1) == 0);
	
	static const bool rval = _get_conv_mm_xs_inner2_has_remainder2<nextlevel_has_remainder, even>::rval;
};

template <>
struct _get_conv_mm_xs_inner2_has_remainder<0>
{
	static const bool rval = true;
};

///#######################################################################

template <int size, typename functor, typename blocktype = _ss_block, bool nextlevel_has_remainder = _get_conv_mm_xs_inner2_has_remainder<size / 2>::rval, bool even = ((size & 0x1) == 0)>
struct _conv_mm_xs_inner2;


template <int size, typename functor, typename blocktype>
struct _conv_mm_xs_inner2<size, functor, blocktype, false, true> 
{	
	static const bool has_remainder						= false;

	static inline void apply2(blocktype const (&a_in)[size], blocktype (&a_out)[1]/*, blocktype &UNUSED(remainder)*/)
	{
		blocktype a_temp[size / 2];
		_conv_mm_xs_worker<size, functor, blocktype>::apply(a_in, a_temp);					///reduce to half
		
		//blocktype dummy;
		_conv_mm_xs_inner2<size / 2, functor, blocktype>::apply2(a_temp, a_out/*, dummy*/);				///reduce again if appropriate
	};
};

template <int size, typename functor, typename blocktype>
struct _conv_mm_xs_inner2<size, functor, blocktype, false, false> 
{	
	static const int even_size							= (size - (size & 0x1));
	typedef blocktype									even_a[even_size];
	static const bool has_remainder						= true;
	
	static inline /*void*/ const blocktype& apply2(blocktype const (&a_in)[size], blocktype (&a_out)[1]/*, blocktype &remainder*/)
	{
		even_a const &a_in2 = *(even_a *)(&a_in);									///extract even part
		blocktype a_temp[even_size / 2];
		_conv_mm_xs_worker<even_size, functor, blocktype>::apply(a_in2, a_temp);				///reduce to half
		
		//blocktype dummy;
		_conv_mm_xs_inner2<size / 2, functor, blocktype>::apply2(a_temp, a_out/*, dummy*/);			///reduce again if appropriate
		return a_in[size - 1];
	};
};

template <int size, typename functor, typename blocktype>
struct _conv_mm_xs_inner2<size, functor, blocktype, true, true> 
{	
	static const bool has_remainder						= true;
	static const bool has_apply3						= true;

	///cannot return reference, because remainder is taken from local variable a_temp
	static inline void /*const blocktype&*/ apply3(blocktype const (&a_in)[size], blocktype (&a_out)[1], blocktype &remainder)
	{
		blocktype a_temp[size / 2];
		_conv_mm_xs_worker<size, functor, blocktype>::apply(a_in, a_temp);					///reduce to half
		
		__if_not_exists(__exists<_conv_mm_xs_inner2<size / 2, functor, blocktype>::has_apply3>)
		{
			blocktype const& remainder1 = _conv_mm_xs_inner2<size / 2, functor, blocktype>::apply2(a_temp, a_out/*, remainder*/);				///reduce again if appropriate
			copyblock<blocktype>::go(remainder, remainder1);					///cannot return reference, because remainder is taken from local variable a_temp
		}
		__if_exists(__exists<_conv_mm_xs_inner2<size / 2, functor, blocktype>::has_apply3>)
		{
			_conv_mm_xs_inner2<size / 2, functor, blocktype>::apply3(a_temp, a_out, remainder);
		}			
	};
	
	template <typename blocktype>
	struct copyblock;
	
	template <>
	struct copyblock<_ss_block>
	{
		static inline void go(blocktype &dest, blocktype const &src)
		{
			dest = src;
		};	
	};

	template <>
	struct copyblock<_ps_block>
	{
		static inline void go(blocktype &dest, blocktype const &src)				///MSVC does not allow direct copy of arrays
		{
			/*
			dest[0] = src[0];
			dest[1] = src[1];
			dest[2] = src[2];
			dest[3] = src[3];
			*/
			
			_mm_store_ps(&dest[0], _mm_load_ps(&src[0]));
		};	
	};	
};

template <int size, typename functor, typename blocktype>
struct _conv_mm_xs_inner2<size, functor, blocktype, true, false> 
{	
	static const int even_size							= (size - (size & 0x1));
	typedef blocktype									even_a[even_size/* + 1*/];
	static const bool has_remainder						= false;
	
	//typedef blocktype									temp2T[size / 2];
	
	//static inline void apply(blocktype const (&a_in)[size], blocktype (&a_out)[size / 2 + 1], blocktype &UNUSED(remainder))
	static inline void apply2(blocktype const (&a_in)[size], blocktype (&a_out)[1]/*, blocktype &UNUSED(remainder)*/)
	{
		even_a const &a_in2 = *(even_a *)(&a_in);									///extract even part
		blocktype a_temp[size / 2];
		_conv_mm_xs_worker<even_size, functor, blocktype>::apply(a_in2, a_temp);			///reduce to half
		
		
		blocktype a_out1[1];
		
		__if_exists(__exists<_conv_mm_xs_inner2<size / 2, functor, blocktype>::has_apply3>)
		{
			blocktype remainder2;
			_conv_mm_xs_inner2<size / 2, functor, blocktype>::apply3(a_temp, a_out1, remainder2);				///reduce again if appropriate
		}
		__if_not_exists(__exists<_conv_mm_xs_inner2<size / 2, functor, blocktype>::has_apply3>)
		{
			blocktype const &remainder2 = _conv_mm_xs_inner2<size / 2, functor, blocktype>::apply2(a_temp, a_out1/*, remainder2*/);
		}
		
		
		blocktype const &remainder1 = a_in[size - 1];		
		blocktype remainderx;
		_mm_xs_worker1<functor, blocktype>::apply(remainderx, remainder1, remainder2);		///operate on odd element
		_mm_xs_worker1<functor, blocktype>::apply(a_out[0], remainderx, a_out1[0]);
	};	 
};
	
///#######################################################################

template <typename functor, typename blocktype>
struct _conv_mm_xs_inner2<2, functor, blocktype, false, true> 
{	
	static const bool has_remainder						= false;

	static inline void apply2(blocktype const (&a_in)[2], blocktype (&a_out)[1]/*, blocktype &UNUSED(remainder)*/)
	{
		_conv_mm_xs_worker<2, functor, blocktype>::apply(a_in, a_out);
	};
};

template <typename functor, typename blocktype>
struct _conv_mm_xs_inner2<3, functor, blocktype, false, false> 
{	
	static const bool has_remainder						= true;
	typedef blocktype									even_a[2];

	static inline /*void*/ const blocktype& apply2(blocktype const (&a_in)[3], blocktype (&a_out)[1]/*, blocktype &remainder*/)
	{
		even_a &a_in2 = *(even_a *)(&a_in);									///extract even part
		_conv_mm_xs_worker<2, functor, blocktype>::apply(a_in2, a_out);						///reduce to half
		
		//remainder = *(blocktype *)(&a_in[2]);
		return a_in[2];
	};
};

///#######################################################################

template 
<
	int size, 
	typename functor, 
	typename blocktype = _ss_block, 
	bool has_remainder = _conv_mm_xs_inner2<size, functor>::has_remainder, 
	bool even = ((size & 0x1) == 0)
>
struct _conv_mm_xs_inner;


template <int size, typename functor, typename blocktype, bool even>
struct _conv_mm_xs_inner<size, functor, blocktype, false, even> 
{	
	static inline void apply(blocktype const (&a_in)[size], blocktype (&a_out)[1])
	{
		//blocktype dummy;
		_conv_mm_xs_inner2<size, functor, blocktype>::apply2(a_in, a_out/*, dummy*/);
	};
};
 
template <int size, typename functor, typename blocktype/*, bool even*/>
struct _conv_mm_xs_inner<size, functor, blocktype, true, true> 
{	
	static inline void apply(blocktype const (&a_in)[size], blocktype (&a_out)[1])
	{
		//blocktype const &remainder = 
		blocktype remainder;
		_conv_mm_xs_inner2<size, functor, blocktype>::apply3(a_in, a_out, remainder);
		
		blocktype const &temp = a_out[0];
		_mm_xs_worker1<functor, blocktype>::apply(a_out[0], temp, remainder);		///operate on odd element
	};
};

template <int size, typename functor, typename blocktype>
struct _conv_mm_xs_inner<size, functor, blocktype, true, false> 
{	
	static inline void apply(blocktype const (&a_in)[size], blocktype (&a_out)[1])
	{
		blocktype const &remainder = _conv_mm_xs_inner2<size, functor, blocktype>::apply2(a_in, a_out);
		
		blocktype const &temp = a_out[0];
		_mm_xs_worker1<functor, blocktype>::apply(a_out[0], temp, remainder);		///operate on odd element
	};
};

template <typename functor, typename blocktype, bool even>
struct _conv_mm_xs_inner<1, functor, blocktype, false, even> 
{	
	static inline void apply(blocktype const (&a_in)[1], blocktype (&a_out)[1])
	{
		a_out[0] = a_in[0];
	};
};

///#######################################################################
///
///#######################################################################

template <int size, typename functor, typename blocktype = _ss_block>
struct _conv_mm_xs2 
{
	typedef blocktype			a1T[1];

	static inline void apply(blocktype const (&a_in)[size], blocktype &a_out)
	{	
		a1T &a_temp = *(a1T *)(&a_out);
		apply(a_in, a_temp);	
	}
	
	static inline void apply(blocktype const (&a_in)[size], blocktype (&a_out)[1])
	{	
		_conv_mm_xs_inner<size, functor, blocktype>::apply(a_in, a_out);	
	}
};

///#######################################################################
///
///#######################################################################

template <int size, typename functor>
struct _mm_ps10 
{		 	
	typedef _ps_block			array_T[size / 4];

	static inline void apply(float const (&a_in)[size], _ps_block &a_out)
	{	
		array_T &a_temp = *(array_T *)(&a_in);	
		_conv_mm_xs2<size / 4, functor, _ps_block>::apply(a_temp, a_out);		
	}

	static inline void apply(float const (&a_in)[size], _ps_block (&a_out)[1])
	{	
		array_T &a_temp = *(array_T *)(&a_in);	
		_conv_mm_xs2<size / 4, functor, _ps_block>::apply(a_temp, a_out);		
	}
};

///#######################################################################

template <int size, typename functor>
struct _mm_ss10
{
	static inline void apply(float const (&a_in)[size], float &a_out)
	{	
		_conv_mm_xs2<size, functor, _ss_block>::apply(a_in, a_out);	
	}

	static inline void apply(float const (&a_in)[size], float (&a_out)[1])
	{	
		_conv_mm_xs2<size, functor, _ss_block>::apply(a_in, a_out);	
	}
};

///#######################################################################
///
///#######################################################################

template 
<
	int size, 
	typename functor, 
	int remainder4 = size % 4, 
	bool lessthan8 = (size < 8), 
	bool has_ss_functor = functor::has_ss_functor
>
struct _conv_mm_xs_inner3;


template <int size, typename functor>
struct _conv_mm_xs_inner3<size, functor, 0, false, true> 
{
	static inline void apply(float const (&a_in)[size], float (&a_out)[1])
	{	
		_ps_block a_temp4;
		_mm_ps10<size, functor::ps_functor>::apply(a_in, a_temp4);
		_mm_ss10<4, functor::ss_functor>::apply(a_temp4, a_out);
	}
};

template <int size, typename functor>
struct _conv_mm_xs_inner3<size, functor, 1, false, true> 
{
	typedef float a4T[size - 1];

	static inline void apply(float const (&a_in)[size], float (&a_out)[1])
	{	
		a4T const &a_in4 = *(a4T *)(&a_in);									///extract 4-even part
		_ps_block a_temp4;
		float a_temp0;
		_mm_ps10<size - 1, functor::ps_functor>::apply(a_in4, a_temp4);	
		_mm_ss10<4, functor::ss_functor>::apply(a_temp4, a_temp0);
		
		_mm_xs_worker1<functor::ss_functor>::apply(a_out[0], a_temp0, a_in[size - 1]);
	}
};

template <int size, typename functor>
struct _conv_mm_xs_inner3<size, functor, 2, false, true> 
{
	typedef float a4T[size - 2];

	static inline void apply(float const (&a_in)[size], float (&a_out)[1])
	{	
		a4T const &a_in4 = *(a4T *)(&a_in);		
		_ps_block a_temp4;
		float a_temp0;							///extract 4-even part
		_mm_ps10<size - 2, functor::ps_functor>::apply(a_in4, a_temp4);	
		_mm_ss10<4, functor::ss_functor>::apply(a_temp4, a_temp0);
		
		float a_temp1;
		_mm_xs_worker1<functor::ss_functor>::apply(a_temp1, a_in[size - 1], a_in[size - 2]);
		_mm_xs_worker1<functor::ss_functor>::apply(a_out[0], a_temp0, a_temp1);
	}
};

template <int size, typename functor>
struct _conv_mm_xs_inner3<size, functor, 3, false, true> 
{
	typedef float a4T[size - 3];

	static inline void apply(float const (&a_in)[size], float (&a_out)[1])
	{	
		a4T const &a_in4 = *(a4T *)(&a_in);									///extract 4-even part
		_ps_block a_temp4;
		float a_temp0;
		_mm_ps10<size - 3, functor::ps_functor>::apply(a_in4, a_temp4);
		_mm_ss10<4, functor::ss_functor>::apply(a_temp4, a_temp0);	
		
		float a_temp1, a_temp2;
		_mm_xs_worker1<functor::ss_functor>::apply(a_temp1, a_in[size - 1], a_in[size - 2]);
		_mm_xs_worker1<functor::ss_functor>::apply(a_temp2, a_in[size - 3], a_temp0);
		_mm_xs_worker1<functor::ss_functor>::apply(a_out[0], a_temp2, a_temp1);
	}
};

///#######################################################################

template <typename functor>																					  ///we need two instances to give compiler most matching partial template specialisation
struct _conv_mm_xs_inner3<1, functor, /*remainder4 = */1, /*lessthan8 =*/true, /*has_ss_functor =*/false> 
{
	static inline void apply(float const (&a_in)[1], float (&a_out)[1])
	{			
		a_out[0] = a_in[0];
	}
};

template <typename functor>																					   ///we need two instances to give compiler most matching partial template specialisation
struct _conv_mm_xs_inner3<1, functor, /*remainder4 = */1, /*lessthan8 =*/true, /*has_ss_functor =*/true> 
{
	static inline void apply(float const (&a_in)[1], float (&a_out)[1])
	{			
		a_out[0] = a_in[0];
	}
};

template <int size, typename functor, int remainder4>
struct _conv_mm_xs_inner3<size, functor, remainder4, /*lessthan8 =*/true, /*has_ss_functor=*/true>			/// has_ss_functor
{
	static inline void apply(float const (&a_in)[size], float (&a_out)[1])
	{			
		_mm_ss10<size, functor::ss_functor>::apply(a_in, a_out);
	}
};
/*
template <int size, typename functor, int remainder4>
struct _conv_mm_xs_inner3<size, functor, remainder4, true, false>			/// !has_ss_functor
{
	static inline void apply(float const (&a_in)[size], float (&a_out)[1])
	{			
		float a_in2[8];
		_mm_ss10<size, functor::ss_functor>::apply(a_in, a_out);
	}
};*/

///#######################################################################
///
///#######################################################################

template <int size, typename functor>						  ///TODO: use _mm_movemask_ps intrinsic where possible. Also, SSE3 has horisontal instructions: http://216.239.59.104/search?q=cache:HmsQIU-m2toJ:www2.technobabble.com.au/article280.html+athlon+64+float+units&hl=en&ct=clnk&cd=22 - haddps, hsubps, haddpd, hsubpd
struct _conv_mm_xs 
{
	typedef float			a1T[1];

	static inline void apply(SSE_ALIGN_FARG float const (&a_in)[size], float &a_out)
	{	
		a1T &a_temp = *(a1T *)(&a_out);
		apply(a_in, a_temp);	
	}
	
	static inline void apply(SSE_ALIGN_FARG float const (&a_in)[size], a1T &a_out)
	{	
		SSE_VALIGN_CHECK(a_in)
	
		_conv_mm_xs_inner3<size, functor>::apply(a_in, a_out);	
	}
};

///#######################################################################
///
///#######################################################################

