/*  ***** BEGIN LICENSE BLOCK *****
 *  Version: RPPL 1.0/GPL 2.0/LGPL 2.1/MPL 1.1/CDDL 1.0
 *  
 *  
 *  The contents of this file are subject to the Roland Pihlakas Public
 *  License Version 1.0 (the "License"); you may not use this file except
 *  in compliance with the License. You may obtain a copy of the License
 *  at http://roland.pri.ee/RPPL/
 *  
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *  
 *  The Original Code is: Modelling of Natural Thinking.
 *  
 *  The Initial Developer of the Original Code is: Roland Pihlakas.
 *  
 *  Portions created by: _____________________________ 
 *  are Copyright (C): _____________________________. 
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 *  
 *  Alternatively, the contents of this file may be used under the terms of 
 *  either the GNU General Public License Version 2 or later (the "GPL"), or 
 *  the GNU Lesser General Public License Version 2.1 or later (the "LGPL"), or 
 *  the Mozilla Public Licence Version 1.1 or later (the "MPL"), or the Common 
 *  Development and Distribution License Version 1.0 or later (the "CDDL"), in 
 *  which case the provisions of the MPL, the GPL, the LGPL or the CDDL are 
 *  applicable instead of those above. If you wish to allow use of your version 
 *  of this file only under the terms of the of either the MPL, the GPL, 
 *  the LGPL or the CDDL, and not to allow others to use your version of this 
 *  file under the RPPL, indicate your decision by deleting the provisions 
 *  above and replace them with the notice and other provisions required by 
 *  the MPL, the GPL, the LGPL or the CDDL. If you do not delete the provisions 
 *  above, a recipient may use your version of this file under the terms of any 
 *  one of the RPPL, the MPL, the GPL, the LGPL or the CDDL.
 *
 *  ***** END LICENSE BLOCK ***** */

//######################################################################################################
//
//######################################################################################################



//#define _tv_vector111_mm_xs2					_vector111_mm_xs_worker

///#######################################################################
///
///#######################################################################

template <int size, typename functor, typename blocktype = _ss_block, int index = 0>
struct _vector111_mm_xs_worker
{
	static inline void apply(blocktype const (&a_in1)[size], blocktype const (&a_in2)[size], blocktype (&a_out)[size])
	{
		_mm_xs_worker1<functor, blocktype>::apply(a_out[index], a_in1[index], a_in2[index]);	
				
		_vector111_mm_xs_worker<size, functor, blocktype, index + 1>::apply(a_in1, a_in2, a_out);
	};
};

template <int size, typename functor, typename blocktype>
struct _vector111_mm_xs_worker<size, functor, blocktype, size>				///end loop
{
	static inline void apply(blocktype const (&UNUSED(a_in1))[size], blocktype const (&UNUSED(a_in2))[size], blocktype (&UNUSED(a_out))[size]) {};
};

///#######################################################################
///
///#######################################################################

template <int size, typename functor>
struct _mm_ps111 
{		 	
	typedef _ps_block			array_T[size / 4];

	static inline void apply(float const (&a_in1)[size], float const (&a_in2)[size], float (&a_out)[size])
	{	
		array_T const &a_in14 = *(array_T *)(&a_in1);	
		array_T const &a_in24 = *(array_T *)(&a_in2);	
		array_T &a_out4 = *(array_T *)(&a_out);	
		
		_vector111_mm_xs_worker<size / 4, functor, _ps_block>::apply(a_in14, a_in24, a_out4);		
	}
};

///#######################################################################

template <int size, typename functor>
struct _mm_ss111
{
	static inline void apply(float const (&a_in1)[size], float const (&a_in2)[size], float (&a_out)[size])
	{	
		_vector111_mm_xs_worker<size, functor, _ss_block>::apply(a_in1, a_in2, a_out);	
	}
};

///#######################################################################
///
///#######################################################################

template <int size, typename functor, int remainder4 = size % 4, bool lessthan4 = (size < 4)>
struct _vector111_mm_xs_inner3;


template <int size, typename functor>
struct _vector111_mm_xs_inner3<size, functor, 0, false> 
{
	static inline void apply(float const (&a_in1)[size], float const (&a_in2)[size], float (&a_out)[size])
	{	
		_mm_ps111<size, functor::ps_functor>::apply(a_in1, a_in2, a_out);
	}
};

template <int size, typename functor>
struct _vector111_mm_xs_inner3<size, functor, 1, false> 
{
	typedef float a4T[size - 1];

	static inline void apply(float const (&a_in1)[size], float const (&a_in2)[size], float (&a_out)[size])
	{	
		a4T const &a_in14 = *(a4T *)(&a_in1);									///extract 4-even part
		a4T const &a_in24 = *(a4T *)(&a_in2);									///extract 4-even part
		a4T &a_out4 = *(a4T *)(&a_out);									///extract 4-even part
		_mm_ps111<size - 1, functor::ps_functor>::apply(a_in14, a_in24, a_out4);
		
		_mm_xs_worker1b<functor::ss_functor>::apply(a_in1[size - 1], a_in2[size - 1], a_out[size - 1]);
	}
};

template <int size, typename functor>
struct _vector111_mm_xs_inner3<size, functor, 2, false> 
{
	typedef float a4T[size - 2];

	static inline void apply(float const (&a_in1)[size], float const (&a_in2)[size], float (&a_out)[size])
	{	
		a4T const &a_in14 = *(a4T *)(&a_in1);									///extract 4-even part
		a4T const &a_in24 = *(a4T *)(&a_in2);									///extract 4-even part
		a4T &a_out4 = *(a4T *)(&a_out);									///extract 4-even part
		_mm_ps111<size - 2, functor::ps_functor>::apply(a_in14, a_in24, a_out4);
		
		_mm_xs_worker1b<functor::ss_functor>::apply(a_in1[size - 2], a_in2[size - 2], a_out[size - 2]);
		_mm_xs_worker1b<functor::ss_functor>::apply(a_in1[size - 1], a_in2[size - 1], a_out[size - 1]);
	}
};

template <int size, typename functor>
struct _vector111_mm_xs_inner3<size, functor, 3, false> 
{
	typedef float a4T[size - 3];

	static inline void apply(float const (&a_in1)[size], float const (&a_in2)[size], float (&a_out)[size])
	{	
		a4T const &a_in14 = *(a4T *)(&a_in1);									///extract 4-even part
		a4T const &a_in24 = *(a4T *)(&a_in2);									///extract 4-even part
		a4T &a_out4 = *(a4T *)(&a_out);									///extract 4-even part
		_mm_ps111<size - 3, functor::ps_functor>::apply(a_in14, a_in24, a_out4);
		
		_mm_xs_worker1b<functor::ss_functor>::apply(a_in1[size - 3], a_in2[size - 3], a_out[size - 3]);
		_mm_xs_worker1b<functor::ss_functor>::apply(a_in1[size - 2], a_in2[size - 2], a_out[size - 2]);
		_mm_xs_worker1b<functor::ss_functor>::apply(a_in1[size - 1], a_in2[size - 1], a_out[size - 1]);
	}
};

///#######################################################################

template <int size, typename functor, int remainder4>
struct _vector111_mm_xs_inner3<size, functor, remainder4, true> 
{
	static inline void apply(float const (&a_in1)[size], float const (&a_in2)[size], float (&a_out)[size])
	{			
		_mm_ss111<size, functor::ss_functor>::apply(a_in1, a_in2, a_out);
	}
};

///#######################################################################
///
///#######################################################################

template <int size, typename functor>
struct _vector111_mm_xs 
{
	typedef float			a1T[1];

	static inline void apply(SSE_ALIGN_FARG float const (&a_in1)[size], SSE_ALIGN_FARG float const (&a_in2)[size], SSE_ALIGN_FARG float (&a_out)[size])
	{	
		SSE_VALIGN_CHECK(a_in1)
		SSE_VALIGN_CHECK(a_in2)
		SSE_VALIGN_CHECK(a_out)
	
		_vector111_mm_xs_inner3<size, functor>::apply(a_in1, a_in2, a_out);	
	}
};

///#######################################################################
///
///#######################################################################
