/*  ***** BEGIN LICENSE BLOCK *****
 *  Version: RPPL 1.0/GPL 2.0/LGPL 2.1/MPL 1.1/CDDL 1.0
 *  
 *  
 *  The contents of this file are subject to the Roland Pihlakas Public
 *  License Version 1.0 (the "License"); you may not use this file except
 *  in compliance with the License. You may obtain a copy of the License
 *  at http://roland.pri.ee/RPPL/
 *  
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *  
 *  The Original Code is: Modelling of Natural Thinking.
 *  
 *  The Initial Developer of the Original Code is: Roland Pihlakas.
 *  
 *  Portions created by: _____________________________ 
 *  are Copyright (C): _____________________________. 
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 *  
 *  Alternatively, the contents of this file may be used under the terms of 
 *  either the GNU General Public License Version 2 or later (the "GPL"), or 
 *  the GNU Lesser General Public License Version 2.1 or later (the "LGPL"), or 
 *  the Mozilla Public Licence Version 1.1 or later (the "MPL"), or the Common 
 *  Development and Distribution License Version 1.0 or later (the "CDDL"), in 
 *  which case the provisions of the MPL, the GPL, the LGPL or the CDDL are 
 *  applicable instead of those above. If you wish to allow use of your version 
 *  of this file only under the terms of the of either the MPL, the GPL, 
 *  the LGPL or the CDDL, and not to allow others to use your version of this 
 *  file under the RPPL, indicate your decision by deleting the provisions 
 *  above and replace them with the notice and other provisions required by 
 *  the MPL, the GPL, the LGPL or the CDDL. If you do not delete the provisions 
 *  above, a recipient may use your version of this file under the terms of any 
 *  one of the RPPL, the MPL, the GPL, the LGPL or the CDDL.
 *
 *  ***** END LICENSE BLOCK ***** */

//######################################################################################################
//
//######################################################################################################



///#######################################################################
///
///#######################################################################

template 
<
	int size1, 
	int size2, 
	bool rotation_unimportant, 
	typename functor, 
	bool a1_divides_4			= ((size1 % 4) == 0), 
	bool a2_divides_4			= ((size2 % 4) == 0), 
	bool both_divide_by_4		= (a1_divides_4 && a2_divides_4)
>
struct _vector11m_mm_xs_inner3; 

///#######################################################################

//template <int size1, int size2, typename functor>
//struct _vector11m_mm_xs_inner3<size1, size2, true, functor, true, false>
template <int size1, int size2, typename functor, bool a2_divides_4>
struct _vector11m_mm_xs_inner3<size1, size2, true, functor, true, a2_divides_4, false>			///if a1 divides by 4, but not both
{
	struct loop_functor
	{
		static inline void go(int i, float const (&a_in1)[size1], float const (&a_in2)[size2], aligned_array<size2> (&a_out)[size1])
		{
			float (&sub_a_out)[size1]	= a_out[i];
			float a2_val				= a_in2[i];
			
			_vector101_mm_xs<size1, functor>::apply(a_in1, a2_val, sub_a_out);
		};
	};

	static inline void apply(float const (&a_in1)[size1], float const (&a_in2)[size2], aligned_array<size2> (&a_out)[size1])
	{			
		_loop_unroller<size2, loop_functor>::apply(a_in1, a_in2, a_out);
	}
};

///#######################################################################

//template <int size1, int size2, typename functor>
//struct _vector11m_mm_xs_inner3<size1, size2, true, functor, false, true>
template <int size1, int size2, bool rotation_unimportant, typename functor, bool a1_divides_4, bool a2_divides_4, bool both_divide_by_4>
struct _vector11m_mm_xs_inner3<size1, size2, rotation_unimportant, functor, a1_divides_4, a2_divides_4, both_divide_by_4>			///if a2 divides by 4, both divide by 4 or none divides by 4
{
	struct loop_functor
	{
		static inline void go(int i, float const (&a_in1)[size1], float const (&a_in2)[size2], aligned_array<size2> (&a_out)[size1])
		{
			float (&sub_a_out)[size2]	= a_out[i];
			float a1_val				= a_in1[i];
			
			_vector101_mm_xs<size2, functor>::apply(a_in2, a1_val, sub_a_out);
		};
	};

	static inline void apply(float const (&a_in1)[size1], float const (&a_in2)[size2], aligned_array<size2> (&a_out)[size1])
	{			
		_loop_unroller<size1, loop_functor>::apply(a_in1, a_in2, a_out);
	}
};

///#######################################################################
///
///#######################################################################

template <int size1, int size2, bool rotation_unimportant, typename functor>
struct _vector11m_mm_xs 
{
	typedef float			a1T[1];

	static inline void apply(SSE_ALIGN_FARG float const (&a_in1)[size1], SSE_ALIGN_FARG float const (&a_in2)[size2], SSE_ALIGN_FARG aligned_array<size2> (&a_out)[size1])		 ///size1 will be first index this way
	{	
		SSE_VALIGN_CHECK(a_in1)
		SSE_VALIGN_CHECK(a_in2)
		SSE_VALIGN_CHECK(a_out)
		
		IFC_S(size1 > 1)
		{
			SSE_VALIGN_CHECK(a_out[1])
		}
		
		_vector11m_mm_xs_inner3<size1, size2, rotation_unimportant, functor>::apply(a_in1, a_in2, a_out);	
	}
};

///#######################################################################
///
///#######################################################################
